# Builtin modules
import logging, unittest
# Own modules
# Local modules
# Program
from .client import Client, Request, FEIterator

class ClientTest(unittest.TestCase):
	def test1(self):
		fec = Client(log=logging.getLogger("test1.Client"))
		self.assertIs(type(fec), Client)
		with fec as c:
			obj = c.request("ping")
			self.assertIs(type(obj), Request)
			self.assertEqual(obj.isSuccess(), True)
			self.assertIs(type(obj.get()), int)
	def test2(self):
		fec = Client(log=logging.getLogger("test2.Client"))
		self.assertIs(type(fec), Client)
		obj = fec.request("ping")
		self.assertIs(type(obj), Request)
		self.assertEqual(obj.isSuccess(), True)
		self.assertIs(type(obj.get()), int)
	def test3(self):
		fec = Client(log=logging.getLogger("test3.Client"))
		objs = fec.request("ping"), fec.request("ping")
		fec.socket.close()
		for obj in objs:
			self.assertIs(type(obj.get()), int)
	def test4(self):
		from itertools import islice
		fec = Client(log=logging.getLogger("test4.Client"))
		inputAddresses = [
			"0x04222cdcba67350ae6ccf038dfb335648bb439d64b212dd1cd3a5ee112ff76de25d026214c36877577fd9eb4fafa6d52100367256f54157be4b339a8f49837206e",
			"0x04da43d782c6506b926e9f5f7627ff62e8fc1e3c55ac5d22729a600743cfb98e0c2376ccd2d636ceb45a5a8630d31ff14253a60ffd2aabce6036b766ac73d4013b",
			"0x047e3d8d5a05808cb46089dc7cc78633b5b2f8760f9a8c9a54736980348c8a2b05f72840038f8475178e4265ae30c1b1f25a3d3c7806c18cea38111dc9ba596cd5",
			"0x04c3f849fb3dcc29a4c1b02a6003ae68f9536182ad3a027f18548754e1c4eac3a31d3fb4f7ce5eb129d78bbe04d05bfc52599ff756c858f50e3bfc0c11074842c2",
			"0x044889c277d5191153701e0e686d6388c7e427087f716b079d8697583cba7cd6c66ca44791286cc5a69eb6b61ef9fb88a4315ba6bac75d17ec114d2168bedebda7",
			"0x04b5cc3e2c3ab73de6445c9adbe67e189a3acaa95aa0a328705c42d4cff747b0b2e3f63ad0ff231d7598d98a0b6d5ed199ec2831928bff1e9ee59de7dea9b8f6a5",
			"0x04da40df988e0b0f0dff125cce7b7a120b870641fa95147d96513334111c852b5c1c36b5661c56034a00fba43c23112f309016fd08a572a9cd8e15d401f2732461",
			"0x045d4944869283654e775307eb4ddfdc0d261395a4bf9c76916bc09f2b2e534a7051331a433c0eb6db01448fb1c752b69a80c79e4ab4fc2976810b57e1e4660b7a",
			"0x045af7f44a5bb9938dbeb5130e8c6e42fbb3196f4fcec508bce8d5eae3404c1ae65a6d2522c53d9d67d4b6d315b15ae2a13c8be985afb3c3113c821723075c48d7",
			"0x0480dc2647ef1aed071be6092c16c7f936f04008c3d4ccedd6ef4b380058cec5e1e6890d7a6c542604ec72422612d82b055e149e3e28c10335576ca2e578a15268",
		]
		it = fec.createIterator("iterTransactionInputs", "btc", "0x8897ea9ceaf18a546cdc513b9179bae31a462ee5bf47818eb7ba909082d11777", chunks=2)
		self.assertIs(type(it), FEIterator)
		for i, (key, data) in islice(enumerate(it), 3):
			self.assertEqual(inputAddresses[i], data["address"])
		for i, (key, data) in islice(enumerate(it, 3), 1):
			self.assertEqual(inputAddresses[i], data["address"])
		it = fec.createIterator("iterTransactionInputs", "btc", "0x8897ea9ceaf18a546cdc513b9179bae31a462ee5bf47818eb7ba909082d11777", fromKey=key, chunks=2)
		for i, (key, data) in enumerate(it, 4):
			self.assertEqual(inputAddresses[i], data["address"])
	def test5(self):
		class Dummy:
			def do(self, c):
				for i in range(5):
					c.request("ping").get()
		from itertools import islice
		fec = Client(log=logging.getLogger("test5.Client"))
		objs = [ fec.request("ping") for i in range(5) ]
		for obj in objs:
			obj.get()
		self.assertEqual( 5, len(list(fec.requests.itervaluerefs())) )
		objs = None
		self.assertNotEqual( 5, len(list(fec.requests.itervaluerefs())) )
		fec.clear()
		d = Dummy()
		d.do(fec)
		del d
		self.assertEqual( 0, len(list(fec.requests.itervaluerefs())) )
	def test6(self):
		fec = Client(log=logging.getLogger("test6.Client"), compression=False)
		objs = [fec.request("ping"), fec.request("ping")]
		fec.connect()
		objs.append(fec.request("ping"))
		fec.socket.close()
		for obj in objs:
			self.assertIs(type(obj.get()), int)
	def test7(self):
		fec = Client(log=logging.getLogger("test7.Client"), compression=False)
		objs = [fec.request("ping"), fec.request("ping")]
		fec.connect()
		objs.append(fec.request("ping"))
		for obj in objs:
			self.assertIs(type(obj.get()), int)
	def test8(self):
		fec = Client("a344613fe3d9ea517ffa0e89e645cdbc", "417e9e027bcd7efb89d250a7cbf701b4", log=logging.getLogger("test8.Client"), compression=False)
		obj = fec.request("sleepWell", [1])
		if obj.isSuccess():
			self.assertEqual(obj.get(), True)

logging.basicConfig(format='[%(levelname).3s][%(asctime)s][%(name)s]: %(message)s', level=logging.INFO)
unittest.main(verbosity=0)



__version__ = "0.0.5"
__doc__ = """
Python client for Fusion Explorer API v{}
Copyright (C) 2021 Fusion Solutions KFT <contact@fusionsolutions.io>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/lgpl.txt>.
""".format(__version__)
from .client import Client
from .exceptions import Error, InitializationError, SocketError, ResponseError, ResponseError, StopSignalError
from .utils import StopSignal
__all__ = ( "Client", "StopSignal", "Error", "InitializationError", "SocketError", "ResponseError", "ResponseError",
	"StopSignalError" )



# Builtin modules
import re, select, ssl, json, traceback, errno
from socket import socket
from time import monotonic
from typing import Any, Union, Callable
# Local modules
from . import __version__
from .utils import deflate
from .exceptions import StopSignalError
# Program
MASKS = {
	"R":select.POLLIN | select.POLLPRI | select.POLLERR | select.POLLHUP | select.POLLNVAL,
	"W":select.POLLOUT | select.POLLERR | select.POLLHUP | select.POLLNVAL,
	"RW":select.POLLOUT | select.POLLIN | select.POLLPRI | select.POLLERR | select.POLLHUP | select.POLLNVAL
}
NOT_CONNECTED = 0
CONNECTING = 1
CONNECTED = 2

class ClientSocket:
	__slots__ = ("log", "client", "endpoint", "port", "ssl", "compression", "connTimeout", "dataTimeout", "signal", "headers",
		"poll", "readBuffer", "writeBuffer", "connectionStatus", "timeoutTimer", "sock", "sockFD", "mask", "error", "sslTimer")
	readBuffer:bytes
	writeBuffer:bytes
	connectionStatus:int
	timeoutTimer:float
	sock:Any
	sockFD:Union[None, int]
	mask:str
	sslTimer:float
	def __init__(self, client:Any, endpoint:str, port:int, ssl:bool, compression:bool,
	connTimeout:int, dataTimeout:int, stopSignal:Any):
		self.log = client.log.getChild("socket")
		self.log.info("Initializing..")
		self.client:Any = client
		self.endpoint:str = endpoint
		self.port:int = port
		self.ssl:bool = ssl
		self.compression:bool = compression
		self.connTimeout:int = connTimeout
		self.dataTimeout:int = dataTimeout
		self.signal:Any = stopSignal
		headers = [
			"POST / HTTP/1.1",
			"Host: {}:{}".format(self.endpoint, self.port),
			"User-Agent: Fusion Explorer client {}".format(__version__),
			"Accept: */*",
			"Connection: Keep-Alive",
			"Content-Type: application/json;charset=utf-8",
		]
		if self.compression:
			headers.append("Accept-Encoding: deflate")
		self.headers:str = "\r\n".join(headers)
		self.poll:Any = select.poll()
		self._reset()
		self.error:Union[None, str] = None
		self.log.info("Initialized")
	def _connect(self, initial:bool=False) -> bool:
		if initial:
			self.log.info("Connecting to {}:{} ..".format(self.endpoint, self.port))
		cerr = self.sock.connect_ex((self.endpoint, self.port))
		if cerr in [errno.EAGAIN, errno.EINPROGRESS]:
			return False
		elif cerr in [0, errno.EISCONN]:
			if self.ssl:
				self.sslTimer = monotonic()
				sslCtx = ssl.create_default_context()
				sslCtx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
				sslCtx.set_ciphers(
					":".join([
						"ECDHE-ECDSA-AES256-GCM-SHA384",
						"ECDHE-RSA-AES256-GCM-SHA384",
						"DHE-RSA-AES256-GCM-SHA384",
						"ECDHE-ECDSA-CHACHA20-POLY1305",
						"ECDHE-RSA-CHACHA20-POLY1305",
						"DHE-RSA-CHACHA20-POLY1305",
						"ECDHE-ECDSA-AES256-SHA384",
						"ECDHE-RSA-AES256-SHA384",
						"DHE-RSA-AES256-SHA256",
						"ECDHE-ECDSA-AES256-SHA",
						"ECDHE-RSA-AES256-SHA",
						"DHE-RSA-AES256-SHA",
						"RSA-PSK-AES256-GCM-SHA384",
						"DHE-PSK-AES256-GCM-SHA384",
						"RSA-PSK-CHACHA20-POLY1305",
						"DHE-PSK-CHACHA20-POLY1305",
						"ECDHE-PSK-CHACHA20-POLY1305",
						"AES256-GCM-SHA384",
						"PSK-AES256-GCM-SHA384",
						"PSK-CHACHA20-POLY1305",
						"ECDHE-PSK-AES256-CBC-SHA384",
						"ECDHE-PSK-AES256-CBC-SHA",
						"SRP-RSA-AES-256-CBC-SHA",
						"SRP-AES-256-CBC-SHA",
						"RSA-PSK-AES256-CBC-SHA384",
						"DHE-PSK-AES256-CBC-SHA384",
						"RSA-PSK-AES256-CBC-SHA",
						"DHE-PSK-AES256-CBC-SHA",
						"AES256-SHA",
						"PSK-AES256-CBC-SHA384",
						"PSK-AES256-CBC-SHA"
					])
				)
				self.sock = sslCtx.wrap_socket(self.sock, do_handshake_on_connect=False, server_hostname=self.endpoint)
				self.log.debug("SSL handshake")
			self.connectionStatus = CONNECTING
			self._setMask("RW")
			if not self.ssl:
				self.connectionStatus = CONNECTED
				connectionDelay =  monotonic()-self.timeoutTimer
				self.timeoutTimer = monotonic()
				self.log.info("Connected in {:.3F} sec".format(connectionDelay))
		elif not initial:
			return self._onError("Connection failed {}[{}]".format(errno.errorcode[cerr], cerr))
		return False
	def _createSocket(self) -> None:
		if self.sock:
			return
		self._reset()
		self.error = None
		self.sock = socket()
		self.sock.setblocking(0)
		self.sockFD = self.sock.fileno()
		self.mask = "R"
		self.poll.register(self.sockFD, MASKS[self.mask])
		self.log.info("Socket created: {}".format(self.sockFD))
	def _doSSLHandshake(self) -> bool:
		try:
			self.sock.do_handshake()
		except ssl.SSLWantReadError:
			self._setMask("R")
			return False
		except ssl.SSLWantWriteError:
			self._setMask("W")
			return False
		except Exception as err:
			return self._onError("SSL Handsake error")
		self._setMask("RW")
		connectionDelay =  monotonic()-self.timeoutTimer
		self.log.info("Connected in {:.3F} sec [SSL: {:.3F} sec]".format(connectionDelay, monotonic()-self.sslTimer))
		self.connectionStatus = CONNECTED
		return True
	def _haveRead(self) -> bool:
		data = None
		if self.ssl and self.connectionStatus == CONNECTING:
			return self._doSSLHandshake()
		try:
			data = self.sock.recv(16<<20)
		except ssl.SSLWantReadError:
			self._setMask("R")
			return False
		except ssl.SSLWantWriteError:
			self._setMask("W")
			return False
		except BlockingIOError:
			pass
		except ConnectionRefusedError:
			return self._onError("Connection {}".format("broken" if self.connectionStatus == CONNECTED else "refused"))
		except:
			return self._onError("Unkown error: {}".format(traceback.format_exc()))
		if data:
			self.readBuffer += data
			self.log.debug(data)
			self.log.info("Read {} bytes [{} bytes in buffer]".format(len(data), len(self.readBuffer)))
			if self._parseReadBuffer():
				return True
			self.timeoutTimer = monotonic()
		else:
			return self._onError("Connection broken")
		return False
	def _haveWrite(self) -> bool:
		sentLength = 0
		if self.connectionStatus == NOT_CONNECTED:
			return self._connect()
		if self.connectionStatus == CONNECTING and self.ssl:
			return self._doSSLHandshake()
		if not self.writeBuffer:
			self._setMask("R")
			return False
		try:
			sentLength = self.sock.send(self.writeBuffer)
		except ssl.SSLWantReadError:
			self._setMask("R")
		except ssl.SSLWantWriteError:
			self._setMask("W")
		except BrokenPipeError:
			return self._onError("Connection broken")
		if sentLength:
			self.timeoutTimer = monotonic()
			self.log.debug(self.writeBuffer[:sentLength])
			self.writeBuffer = self.writeBuffer[sentLength:]
			self.log.info("Sent {} bytes [{} still in buffer]".format(sentLength, len(self.writeBuffer)))
		if not self.writeBuffer:
			self._setMask("R")
		return False
	def _onError(self, err:str) -> bool:
		self.log.info(err)
		self.error = err
		return True
	def _parseReadBuffer(self) -> bool:
		endLine = b"\r\n"
		pos = self.readBuffer.find(endLine*2)
		if pos == -1:
			endLine = b"\n"
			pos = self.readBuffer.find(endLine*2)
		elif pos > 4096:
			return self._onError("HTTP headers are too long")
		while pos != -1:
			rawData = self.readBuffer[pos+len(endLine)*2:]
			rawHeaders = self.readBuffer[:pos].decode("ISO-8859-1").split(endLine.decode("ISO-8859-1"))
			if not rawHeaders:
				return self._onError("Invalid HTTP headers")
			httpResponse = rawHeaders.pop(0).split(" ")
			if len(httpResponse) < 2:
				return self._onError("Invalid HTTP response")
			if httpResponse[1] == "503":
				return self._onError("Server offline")
			elif httpResponse[1] != "200":
				return self._onError("Request failure")
			#
			headers = {}
			for rawHeader in rawHeaders:
				s = rawHeader.find(":")
				if s <= 0 or s > 64:
					return self._onError("Header key too long")
				headers[ rawHeader[:s].strip().lower() ] = rawHeader[s+1:].strip()
			cLengthStr = headers.get("content-length", "")
			if not cLengthStr.isdigit():
				return self._onError("Invalid HTTP header value for content-length")
			cLength = int(cLengthStr)
			if cLength < 0:
				return self._onError("Invalid HTTP header value for content-length")
			cEncoding = headers.get("content-encoding", "")
			if cEncoding == "":
				compression = False
			elif cEncoding == "deflate":
				compression = True
			else:
				return self._onError("Request HTTP encoding not supported")
			cType = headers.get("content-type", "")
			if cType == "":
				return self._onError("Invalid HTTP header value for content-type")
			cTypeS = re.findall("charset=([^ ]*)", cType, re.I)
			charset = cTypeS[0] if len(cTypeS) == 1 else "iso-8859-1"
			try:
				"".encode(charset)
			except:
				return self._onError("Not supported charset")
			#
			if cLength != len(rawData):
				return False
			#
			self.readBuffer = self.readBuffer[pos+len(endLine)*2+cLength:]
			try:
				if compression:
					data = json.loads( deflate.decompress(rawData[:cLength]).decode(charset) )
				else:
					data = json.loads( rawData[:cLength].decode(charset) )
			except:
				return self._onError("Invalid response content")
			if type(data) is list:
				for d, uid in zip(data, headers.get("x-jsonrequestid", ","*len(data)).split(",")):
					self.client._gotResult(
						d,
						headers.get("x-connectionid", ""),
						headers.get("x-httprequestid", ""),
						uid
					)
			else:
				self.client._gotResult(
					data,
					headers.get("x-connectionid", ""),
					headers.get("x-httprequestid", ""),
					headers.get("x-jsonrequestid", ""),
				)
			pos = self.readBuffer.find(endLine*2)
		return False
	def _removeMask(self) -> None:
		self.mask = "R"
		try: self.poll.unregister(self.sockFD)
		except: pass
	def _reset(self) -> None:
		self.readBuffer = b""
		self.writeBuffer = b""
		self.connectionStatus = NOT_CONNECTED
		self.timeoutTimer = 0.0
		self.sock = None
		self.sockFD = None
		self.mask = "R"
		self.sslTimer = 0.0
	def _setMask(self, newMask:str) -> None:
		if self.mask != newMask:
			self.poll.modify(self.sockFD, MASKS[newMask])
			self.mask = newMask
	def _write(self, data:bytes) -> None:
		self.writeBuffer += data
		self._setMask("RW")
	def close(self) -> None:
		if self.sock:
			self._removeMask()
			try: self.sock.close()
			except: pass
			self._reset()
			self.log.info("Closed")
	def connect(self) -> None:
		if self.connectionStatus != NOT_CONNECTED:
			return
		self._createSocket()
		self.timeoutTimer = monotonic()
		self._setMask("W")
		self._connect(initial=True)
		self.loop(lambda:not self.isConnected())
	def isAlive(self) -> bool:
		return not (
			(self.connectionStatus == NOT_CONNECTED) or
			(self.connectionStatus == CONNECTING and monotonic()-self.timeoutTimer > self.connTimeout) or
			(self.connectionStatus == CONNECTED and monotonic()-self.timeoutTimer > self.dataTimeout)
		)
	def isConnected(self) -> bool:
		return self.connectionStatus == CONNECTED
	def loop(self, whileFn:Callable=None) -> None:
		checkTimer = monotonic()
		while True and (whileFn is None or whileFn()):
			if self.signal.get():
				raise StopSignalError()
			if monotonic()-checkTimer >= 1:
				checkTimer = monotonic()
				if not self.isAlive():
					self._onError("Timeout")
					return
			for fd, bitmask in self.poll.poll(100):
				if fd != self.sockFD: continue
				readable = bool(bitmask & (select.POLLIN | select.POLLPRI))
				writeable = bool(bitmask & select.POLLOUT)
				error = bool(bitmask & (select.POLLERR | select.POLLHUP | select.POLLNVAL))
				# self.log.debug("Poll result: {}{}{}".format(
				# 	"read " if readable else "",
				# 	"write " if writeable else "",
				# 	"error " if error else "",
				# ))
				isConnected = self.connectionStatus == CONNECTED
				if readable and self._haveRead():
					if isConnected:
						self.close()
					return
				if writeable and self._haveWrite():
					if isConnected:
						self.close()
					return
				if error:
					self.close()
					self._onError("Client socket broken")
					return
	def send(self, data:bytes, auth:str=None) -> None:
		if self.connectionStatus == CONNECTED:
			headers = self.headers
			if auth:
				headers += "\r\nX-Auth: {}".format(auth)
			headers += "\r\nContent-Length: {}\r\n\r\n".format(len(data))
			self._write( headers.encode("iso-8859-1") + data )



# Builtin modules
import zlib, signal, json
from typing import Any, List
# Local modules
# Program
class deflate:
	@staticmethod
	def compress(data:bytes) -> bytes:
		o = zlib.compressobj(9, zlib.DEFLATED, -zlib.MAX_WBITS, zlib.DEF_MEM_LEVEL, 0)
		r = o.compress(data)
		r += o.flush()
		return r
	@staticmethod
	def decompress(data:bytes) -> bytes:
		o = zlib.decompressobj(-zlib.MAX_WBITS)
		r = o.decompress(data)
		r += o.flush()
		return r

class StopSignal:
	_i:bool = False
	killed:bool = False
	@classmethod
	def activate(self) -> Any:
		assert self._i == False, "Aleady initialized"
		self._i = True
		signal.signal(signal.SIGINT, self.kill)
		return self
	@classmethod
	def kill(self, signum:Any=None, frame:Any=None) -> None:
		self.killed = True
	@classmethod
	def get(self) -> bool:
		return self.killed

def hexToBytes(data:str) -> bytes:
	if data[:2].lower() == "0x":
		data = data[2:]
	return bytes.fromhex(data)

def chunks(lst:List[Any], n:int, pl:int) -> List[Any]:
	r = []
	rl = [ json.dumps(d).encode("utf8") for d in lst ]
	sr = []
	srl = 2
	for i in range(len(rl)):
		sr.append(rl[i])
		srl += len(rl[i])
		if len(sr) == n or (len(rl)-1 > i and srl+len(rl[i+1])+len(sr) > pl):
			r.append(b"["+(b",".join(sr))+b"]")
			srl = 2
			sr = []
	if sr:
		r.append(b"["+(b",".join(sr))+b"]")
	return r



# Builtin modules
import json, re, hmac, logging, weakref
from hashlib import sha256
from time import time, sleep
from collections import OrderedDict
from itertools import starmap
from typing import Any, Union, Callable, MutableMapping, List, Dict, Tuple, cast
# Local modules
from .exceptions import *
from .utils import hexToBytes, StopSignal, chunks
from .clientSocket import ClientSocket
# Program
class Client:
	__slots__ = ("log", "pubkey", "seckey", "hexNumbers", "connTimeout", "timeout", "compression", "timeWindow", "retryCount",
		"retryDelay", "stopSignal", "id", "requests", "socket" )
	endpoint:str = "api.fusionexplorer.io"
	port:int = 443
	ssl:bool = True
	max_bulk_request:int = 0xFF
	max_payload:int = 0xFFFF
	pubkey:bytes
	seckey:bytes
	def __init__(self, projectPublicKey:Union[str, bytes]=None, projectSecretKey:Union[str, bytes]=None,
	hexNumbers:bool=False, connectTimeout:int=15, timeout:int=320, compression:bool=True, timeWindow:int=60, retryCount:int=10,
	retryDelay:int=5, stopSignal:Any=None, log:Any=None):
		self.log = log or logging.getLogger("FusionExplorer")
		self.log.info("Initializing..")
		if projectPublicKey is not None and len(projectPublicKey) > 0:
			if projectSecretKey is None:
				raise InitializationError("`projectSecretKey` is required when `projectPublicKey` is set")
			if type(projectPublicKey) is str:
				if not re.match(r'^(0x|)[0-9a-f]{32}$', cast(str, projectPublicKey), re.I):
					raise InitializationError("`projectPublicKey` must be 32 character hex string")
				self.pubkey = hexToBytes(cast(str, projectPublicKey))
			elif type(projectPublicKey) is bytes:
				self.pubkey = cast(bytes, projectPublicKey)
			else:
				raise InitializationError("`projectPublicKey` must be string or bytes")
			if len(self.pubkey) != 16:
				raise InitializationError("`projectPublicKey` must be 16 byte long")
			if type(projectSecretKey) is str:
				if not re.match(r'^(0x|)[0-9a-f]{32}$', cast(str, projectSecretKey), re.I):
					raise InitializationError("`projectSecretKey` must be 32 character hex string")
				self.seckey = hexToBytes(cast(str, projectSecretKey))
			elif type(projectSecretKey) is bytes:
				self.seckey = cast(bytes, projectSecretKey)
			else:
				raise InitializationError("`projectSecretKey` must be string or bytes")
			if len(self.seckey) != 16:
				raise InitializationError("`projectSecretKey` must be 16 byte long")
		else:
			self.pubkey = b""
			self.seckey = b""
		self.hexNumbers = hexNumbers
		self.connTimeout = connectTimeout
		self.timeout = timeout
		self.compression = compression
		self.timeWindow = timeWindow
		self.retryCount = retryCount
		self.retryDelay = retryDelay
		self.stopSignal = stopSignal or StopSignal
		self.id:int = 0
		self.requests:MutableMapping = weakref.WeakValueDictionary()
		self.socket:Any = ClientSocket(self, self.endpoint, self.port, self.ssl, self.compression,
			self.connTimeout, self.timeout, self.stopSignal)
		self.log.debug("Public token: {}".format(self.pubkey.hex()))
		self.log.debug("Secret token: {}...{}".format(self.seckey[:1].hex(), self.seckey[-1:].hex()))
		self.log.info("Initialized")
	def __del__(self) -> None:
		self.close()
	def __getstate__(self) -> Dict[str, Any]:
		return {
			"log":        self.log,
			"pubkey":     self.pubkey,
			"seckey":     self.seckey,
			"hexNumbers": self.hexNumbers,
			"connTimeout":self.connTimeout,
			"timeout":    self.timeout,
			"compression":self.compression,
			"timeWindow": self.timeWindow,
			"retryCount": self.retryCount,
			"retryDelay": self.retryDelay,
			"stopSignal": self.stopSignal,
		}
	def __setstate__(self, states:Dict[str, Any]):
		self.log         =states["log"]
		self.pubkey      =states["pubkey"]
		self.seckey      =states["seckey"]
		self.hexNumbers  =states["hexNumbers"]
		self.connTimeout =states["connTimeout"]
		self.timeout     =states["timeout"]
		self.compression =states["compression"]
		self.timeWindow  =states["timeWindow"]
		self.retryCount  =states["retryCount"]
		self.retryDelay  =states["retryDelay"]
		self.stopSignal  =states["stopSignal"]
		self.killFn      =states["killFn"]
		self.forceKillFd =states["forceKillFn"]
		self.id          =0
		self.requests    =weakref.WeakValueDictionary()
		self.socket      =None
	def __enter__(self) -> Any:
		return self.clone()
	def __exit__(self, type:Any, value:Any, traceback:Any):
		self.close()
	def _checkSocketError(self):
		if self.socket.error:
			err = self.socket.error
			self.socket.close()
			raise SocketError(err)
	def _connect(self, sendOlderRequests:bool=True):
		if not self.socket.isAlive():
			self.socket.connect()
			self._checkSocketError()
			self.log.info("Connected")
			if sendOlderRequests:
				objs = list(filter(lambda x: not x.isDone(), self.requests.values()))
				if objs:
					self.log.info("Sending {} previous requests".format(len(objs)))
					objs.sort(key=lambda x: x._requestTime)
					for chunk in chunks(list(map( lambda x: x._toJson(), objs)), self.max_bulk_request, self.max_payload):
						self.socket.send(chunk, self._createAuth(chunk))
	def _createAuth(self, data:bytes) -> str:
		if not self.seckey: return ""
		ts = int(time()+self.timeWindow).to_bytes(4, "big")
		return b"".join((
			ts,
			self.pubkey,
			hmac.digest(
				self.seckey,
				b"".join([
					ts,
					self.pubkey,
					sha256(data).digest(),
				]),
				"SHA256"
			)
		)).hex()
	def _get(self, id:Any) -> None:
		def wh() -> bool:
			for obj in self.requests.values():
				if obj.getID() == id:
					if not obj.isDone():
						return True
					else:
						return False
			return False
		if id not in self.requests:
			raise ResponseError("Unknown request ID: {}".format(id))
		if not self.socket.isAlive():
			raise SocketError("Not connected")
		self.socket.loop(wh)
		self._checkSocketError()
		if id not in self.requests:
			raise ResponseError("Unknown request ID: {}".format(id))
		elif not self.requests[id].isDone():
			raise ResponseError("Did not get any response for ID: {}".format(id))
	def _gotResult(self, data:Dict[str, Any], connectionID:str, HTTPRequestID:str, JSONRequestID:str) -> None:
		if data['id'] not in self.requests:
			return
		obj = self.requests[data['id']]
		success = not ("error" in data and data["error"])
		self.log.info("Got result for ID: {}".format(data["id"]))
		uid = "-".join((connectionID, HTTPRequestID, JSONRequestID))
		self.log.info("Got result for UID: {}".format(uid))
		obj._parseResponse( uid, data )
	def _tryAgain(self, fn:Callable, *args, **kwargs) -> Any:
		lastErr = None
		c = self.retryCount
		while True:
			try:
				return fn(*args, **kwargs)
			except Exception as err:
				lastErr = err
				if isinstance(err, SocketError):
					self.log.debug("Got error: {}".format(err))
					if c > 0:
						if c != self.retryCount:
							sleep(self.retryDelay)
						c -= 1
						self.connect()
						continue
				break
		if isinstance(lastErr, SocketError):
			self.log.error(lastErr.message)
		raise lastErr from None
	def clear(self) -> None:
		self.requests.clear()
	def clone(self, **kwargs) -> Any:
		opts:Dict[str, Any] = {
			"projectPublicKey":self.pubkey,
			"projectSecretKey":self.seckey,
			"hexNumbers":self.hexNumbers,
			"connectTimeout":self.connTimeout,
			"timeout":self.timeout,
			"compression":self.compression,
			"timeWindow":self.timeWindow,
			"retryCount":self.retryCount,
			"retryDelay":self.retryDelay,
			"log":self.log,
			"stopSignal":self.stopSignal,
		}
		opts.update(kwargs)
		return Client(**opts)
	def close(self) -> None:
		if hasattr(self, "socket"):
			self.socket.close()
			self.clear()
	def connect(self) -> None:
		self.socket.close()
		self._tryAgain(self._connect)
	def createIterator(self, method:str, *args:List[Any],
	sortBy:str=None, fromKey:str=None, desc:bool=None, bitmask:int=None, chunks:int=12) -> Any:
		if not method.startswith("iter"):
			raise InitializationError("For iterating you need choose method which name begins with `iter`.")
		kwargs:Dict[str, Any] = { "limit":chunks }
		if sortBy is not None: kwargs["sortBy"] = sortBy
		if fromKey is not None: kwargs["fromKey"] = fromKey
		if desc is not None: kwargs["desc"] = desc
		if bitmask is not None: kwargs["bitmask"] = bitmask
		return FEIterator(self, method, args, kwargs)
	def request(self, method:str, args:List[Any]=[], kwargs:Dict[str, Any]={}, id:Union[str, int]=None) -> Any:
		if id == None:
			id = self.id
			self.id += 1
		if type(id) not in [int, str]:
			raise RequestError("Request ID can be only str or int")
		if id in self.requests:
			raise RequestError("Request ID already in use: {}".format(id))
		obj = Request(
			[self._tryAgain, self._get],
			id,
			method,
			args,
			kwargs,
			self.hexNumbers
		)
		payload = json.dumps(obj._toJson()).encode("utf8")
		if len(payload) > self.max_payload:
			raise RequestError("Request payload too long")
		self.requests[id] = obj
		self.log.info("Request queued: {} [{}]".format(method, id))
		if self.socket.isAlive():
			self.socket.send(payload, self._createAuth(payload))
		return obj

class Request:
	__slots__ = (
		"_getter", "_id", "_method", "_args", "_kwargs", "_hexNumbers", "_requestTime",
		"_responseTime", "_uid", "_done", "_success", "_response", "__weakref__"
	)
	def __init__(self, getter:Any, id:Any, method:str, args:List[Any], kwargs:Dict[str, Any], hexNumbers:bool):
		self._getter = getter
		self._id = id
		self._method = method
		self._args = args
		self._kwargs = kwargs
		self._hexNumbers = hexNumbers
		self._requestTime:float = time()
		self._responseTime:float = 0.0
		self._uid:str = ""
		self._done:bool = False
		self._success:bool = False
		self._response:Any = None
	def _get(self):
		self._getter[0](self._getter[1], self._id)
	def _parseResponse(self, uid:str, data:Any) -> None:
		self._done = True
		self._responseTime = time()
		self._uid = uid
		self._success = not ("error" in data and data["error"])
		self._response = data["result"] if self._success else data["error"]
	def _toJson(self) -> Dict[str, Any]:
		return {
			"jsonrpc":"fsp1",
			"args":self._args,
			"kwargs":self._kwargs,
			"method":self._method,
			"id":self._id,
			"hexNumbers":self._hexNumbers
		}
	def get(self) -> Any:
		if not self._done:
			self._get()
		return self._response
	def getDelay(self) -> float:
		if not self._done:
			self._get()
		return self._requestTime - self._responseTime
	def getID(self) -> Any:
		return self._id
	def isDone(self) -> bool:
		return self._done
	def getUID(self) -> str:
		if not self._done:
			self._get()
		return self._uid
	def isSuccess(self) -> bool:
		if not self._done:
			self._get()
		return self._success

class FEIterator:
	def __init__(self, client:Any, method:str, args:tuple, kwargs:Dict[str, Any]):
		self.client = client
		self.method = method
		self.args = args
		self.kwargs = kwargs
		self.cache:list = []
	def __enter__(self) -> Any:
		return self
	def __exit__(self, type:Any, value:Any, traceback:Any) -> None:
		pass
	def __iter__(self) -> Any:
		return self
	def __request__(self) -> None:
		if not self.cache:
			req = self.client.request(self.method, self.args, self.kwargs)
			data = req.get()
			if not req.isSuccess():
				raise ResponseError(data)
			if type(data) is list:
				self.cache = [ (x["key"], x["data"]) for x in data ]
	def __next__(self) -> Tuple[Any, Any]:
		self.__request__()
		if not self.cache:
			raise StopIteration
		key, data = self.cache.pop(0)
		self.kwargs["fromKey"] = key
		return key, data
	def checkNext(self) -> Tuple[Any, Any]:
		self.__request__()
		if not self.cache:
			raise StopIteration
		return self.cache[0]
	next = __next__



#!/usr/bin/env python3
import os, shutil
import feClient
try:
	from setuptools import setup
except ImportError:
	from distutils.core import setup

pwd = os.path.abspath(os.path.dirname(__file__))

setup(
	name = "python-feclient",
	version = feClient.__version__,
	description = "Client for Fusion Explorer API",
	keywords = "explorer client fusion solutions fusionsolutions fusionexplorer",
	author = "Andor `iFA88` Rajci - Fusions Solutions KFT",
	author_email = "ifa@fusionsolutions.io",
	url = "https://github.com/FusionSolutions/python-feclient",
	license = "GPL-3",
	# setup_requires=["setuptools>=25"],
	# install_requires=["setuptools>=25"],
	package_dir={"feClient": "feClient"},
	packages=["feClient"],
	long_description=open(os.path.join(pwd, "README.md")).read(),
	long_description_content_type="text/markdown",
	zip_safe=False,
	scripts=["feClient/fexplorer-cli"],
	python_requires=">=3.8.*",
	classifiers=[ # https://pypi.org/pypi?%3Aaction=list_classifiers
		"Development Status :: 4 - Beta",
		"Topic :: Utilities",
		"Programming Language :: Python :: 3 :: Only",
		"Programming Language :: Python :: 3.8",
		"Programming Language :: Python :: 3.9",
		"License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
	],
)